#!/usr/bin/python

## expendable

## Copyright (C) 2008 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

class ModelError(Exception):
    def __init__ (self, model, reason):
        self.model = model
        self.reason = reason

    def __str__ (self):
        return "%s in %s" % (str (self.reason), str (self.model))

class Model:
    def __init__ (self):
        self.models = []
        self.model_data = []
        self.per = 0
        self.parent = None
        self.set_delay (0)

    def set_delay (self, delay):
        self.delay = delay
        if delay > 0:
            # We have to start late.
            self.per = -delay
        else:
            # We have to start early.
            self.run (nper=-delay)

    def add_model (self, model):
        """Add a model to be called each iteration."""
        model.active = True
        self.models.append (model)
        model.parent = self

    def get_models (self):
        return self.models

    def get_model_data (self):
        """Return a list of dicts."""
        return self.model_data

    def format_datasets (self):
        datasets = self.get_datasets ()
        for set in datasets:
            data = set['data']
            data = map (lambda xy: (xy[0] + self.delay, xy[1]), data)
            if self.delay < 0:
                data = data[-self.delay:]
            set['data'] = data
        return datasets

    def get_datasets (self):
        return []

    def run (self, nper=-1):
        """Run a specified number of iterations, or keep running until end."""
        while nper != 0:
            for model in self.models:
                if model.active and model.run (nper=1):
                    model.active = False

            if self.per >= 0:
                if self.iter ():
                    self.per += 1
                    return True

            self.per += 1
            if nper > 0:
                nper -= 1

        return False

    def cash_available (self):
        """Returns the cash available now less the liabilities
        incurred to date."""
        min = 0.0
        cash_available = 0.0
        for model in self.models:
            if model.active == False:
                continue
            avail = model.cash_available ()
            if cash_available + avail < min:
                min = cash_available + avail
            cash_available += avail
        if min < 0:
            cash_available = min
        return cash_available

    def cash_available_next_month (self):
        """Returns the cash that will be available next month less
        the liabilities that will be incurred during the next month."""
        min = 0.0
        cash_available = 0.0
        for model in self.models:
            if model.active == False:
                continue
            avail = model.cash_available_next_month ()
            if cash_available + avail < min:
                min = cash_available + avail
            cash_available += avail
        return min

    def test_adjust (self, amount, by_account):
        raise RuntimeError, "Not implemented"

    def adjust (self, amount, by_account):
        raise RuntimeError, "Not implemented"

    def _adjust_recursive (self, amount, by_account, test=False):
        if amount == 0:
            return 0

        total_adjustment = 0
        if by_account in self.models:
            # One of the accounts linked to this is trying to withdraw
            # money.  If we don't have enough, pass the request on to
            # our own linked account.
            if self.parent != None:
                if test:
                    total_adjustment = self.parent.test_adjust (amount, self)
                else:
                    total_adjustment = self.parent.adjust (amount, self)
                        
        else:
            for model in self.models:
                if test:
                    adjustment = model.test_adjust (amount, self)
                else:
                    try:
                        adjustment = model.adjust (amount, self)
                    except:
                        adjustment = 0

                amount += adjustment
                total_adjustment += adjustment
                if amount == 0:
                    break

        return -total_adjustment

    def iter (self):
        """Run an iteration and append to self.model_data.
        
        Returns True if this iteration is the last."""
        raise RuntimeError, "Not implemented"

class Data:
    def __init__ (self):
        self.accounts = []
        self.name = None
        self.type = None
        self.description = None
        self.delay = 0
        self.plot = True
        self.redisplay_fn = lambda x: False
        self.data = {}

    def _set_last_known_good (self):
        last_known_good = {}
        for key, value in self.data.iteritems ():
            last_known_good[key] = value
        last_known_good['_delay'] = self.delay
        self.last_known_good = last_known_good

    def set_redisplay_fn (self, redisplay_fn):
        self.redisplay_fn = redisplay_fn

    def redisplay (self):
        self.redisplay_fn ()
        self._set_last_known_good ()

    def set_type (self, type):
        self.type = type

    def get_type (self):
        return self.type

    def set_name (self, name):
        self.name = name

    def get_name (self):
        return self.name

    def set_description (self, description):
        self.description = description

    def get_description (self):
        return self.description

    def set_delay (self, delay):
        self.delay = delay

    def get_delay (self):
        return self.delay

    def set_plot (self, plot):
        self.plot = plot

    def get_plot (self):
        return self.plot

    def is_leaf_node (self):
        return False

    def read (self, xml):
        """Parse the XML definition for this account."""
        raise RuntimeError, "Not implemented"

    def write (self):
        """Return a string XML definition for this account."""
        raise RuntimeError, "Not implemented"

    def ui (self, variables_vbox, results_vbox):
        """Fill a VBox with widgets for adjustment, and another for results."""
        return

    def update_results (self, modeldata):
        """Update the results widgets."""
        return

    def destroy (self):
        """Disconnect any signals."""
        return

    def get_model (self):
        """Return a Model for this account."""
        raise RuntimeError, "Not implemented"
