#!/usr/bin/python

## expendable

## Copyright (C) 2008 Tim Waugh <twaugh@redhat.com>

## This program is free software; you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation; either version 2 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program; if not, write to the Free Software
## Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

import os
import sys
if __name__ == '__main__':
    sys.path.append (os.environ.get ('top_srcdir', '.'))

from gettext import gettext as _
import gtk.glade
import locale
import account

name=_("Savings account")
description=_("Instant-access savings account with taxable interest.")

class Model(account.Model):
    def __init__ (self, rate, taxrate, balance):
        account.Model.__init__ (self)
        self.rate = rate
        self.taxrate = taxrate
        self.mrate = pow (1 + self.rate, 1 / 12.0) - 1
        self.amount = balance
        self.balance = balance
        self.gross = 0
        self.net = 0

    def test_adjust (self, amount, by_account):
        return self.adjust (amount, by_account, test=True)

    def adjust (self, amount, by_account, test=False):
        adjustment = amount
        cash_available_next_month = self.cash_available_next_month ()
        if by_account in self.models:
            cash_available_next_month -= by_account.cash_available_next_month ()

        if (amount < 0 and
            self.balance + amount < -cash_available_next_month):
            adjustment = -(self.balance + cash_available_next_month)
            if adjustment > 0:
                adjustment = 0.0

        if not test:
            self.balance += adjustment

        amount -= adjustment
        adjustment += self._adjust_recursive (amount, by_account, test=test)
        return -adjustment

    def cash_available (self):
        cash_available = account.Model.cash_available (self)
        return cash_available + self.balance

    def iter (self):
        gross = int (100 * self.balance * self.mrate) * 0.01

        if self.per == 0:
            if self.delay > 0:
                contrib = self.parent.adjust (-self.amount, self)
                if contrib < self.amount:
                    raise account.ModelError (self,
                                              "Not enough to open account")

            # Balance already set in __init__.
            gross = 0

        tax = int (100 * gross * self.taxrate) * 0.01
        net = gross - tax
        self.balance += net
        self.gross += gross
        self.net += net

        self.model_data.append( { 'per': self.per,
                                  'balance': self.balance,
                                  'totalgross': self.gross,
                                  'totalnet': self.net,
                                  'totaltax': self.gross - self.net,
                                  'gross': gross,
                                  'net': net,
                                  'tax': tax } )

        return False

    def get_datasets (self):
        model_data = self.model_data
        sdata = map (lambda x: (x['per'], x['balance']), model_data)
        return [{'name': self.account.get_name (),
                 'data': sdata}]

class Data(account.Data):
    required_tags = ['rate', 'taxrate']
    glade = "savings.glade"

    def __init__ (self):
        account.Data.__init__ (self)
        for dir in sys.path:
            f = os.path.join (dir, self.glade)
            if os.access (f, os.R_OK):
                gld = gtk.glade.XML (f)
                break

        data = {}
        for name in ['rate', 'taxrate']:
            widget = gld.get_widget (name)
            data[name] = widget.get_value () / 100.0

        for name in ['balance']:
            widget = gld.get_widget (name)
            data[name] = widget.get_value ()

        self.data = data

    def read (self, xml):
        data = {}
        for child in xml.getchildren ():
            # Floats
            if child.tag in ['rate',
                             'taxrate',
                             'balance']:
                data[child.tag] = float (child.text)
            else:
                print "Not understood: %s" % child.tag

        for required in self.required_tags:
            if not data.has_key (required):
                raise RuntimeError, "Required tags: %s" % self.required_tags

        self.data = data
        self._set_last_known_good ()

    def write (self):
        xml = "<savings>\n"
        for var in self.required_tags:
            value = self.data[var]
            if type (value) == float:
                value = "%.4f" % value
            xml += "  <%s>%s</%s>\n" % (var, value, var)
        for optvar in ['balance']:
            v = self.data.get (optvar, None)
            if v != None:
                if type (v) == float:
                    v = "%.4f" % v
                xml += "  <%s>%s</%s>\n" % (optvar, v, optvar)
        xml += "</savings>\n"
        return xml

    def ui (self, variables_vbox, results_vbox):
        for dir in sys.path:
            f = os.path.join (dir, self.glade)
            if os.access (f, os.R_OK):
                gld = gtk.glade.XML (f)
                break

        w = gld.get_widget ('variables_table')
        w.reparent (variables_vbox)
        widget = gld.get_widget ('balance')
        widget.set_value (self.data['balance'])
        widget.connect ("value-changed", self.variable_changed, 'balance')
        for var in ['rate', 'taxrate']:
            widget = gld.get_widget (var)
            widget.set_value (self.data[var] * 100.0)
            widget.connect ("value-changed", self.rate_changed, var)

        w = gld.get_widget ('results_hbox')
        w.reparent (results_vbox)
        self.result_widget = {}
        for name in ['fbalance', 'totalnet']:
            self.result_widget[name] = gld.get_widget (name)

    def rate_changed (self, widget, name):
        self.data[name] = widget.get_value () / 100.0
        try:
            self.redisplay ()
        except:
            widget.set_value (self.last_known_good[name] * 100.0)

    def variable_changed (self, widget, name):
        self.data[name] = widget.get_value ()
        try:
            self.redisplay ()
        except:
            widget.set_value (self.last_known_good[name])

    def update_results (self, modeldata):
        fbalance = self.result_widget['fbalance']
        totalnet = self.result_widget['totalnet']
        n = len (modeldata)
        if n > 0:
            last = modeldata[n - 1]
            fbalance.set_text (locale.currency (last['balance'], True, True))
            totalnet.set_text (locale.currency (last['totalnet'], True, True))
        else:
            fbalance.set_text ('')
            totalnet.set_text ('')

    def get_model (self):
        return Model (**self.data)

if __name__ == '__main__':
    from pprint import *
    from xml.etree.ElementTree import XML
    d = Data ()
    w = d.write ()
    xml = XML (w)
    d.read (xml)

    s = Model (rate=0.065,
               taxrate=0.0,
               balance=1000)
    s.run (nper=13)
    data = s.get_model_data ()
    #pprint (data)
    if abs (data[12]['balance'] - 1064.93) > 0.01:
        raise RuntimeError, "Savings account incorrect interest"

    import mortgage
    m = mortgage.Model (rate=0.06,
                        loan_amount=20000,
                        nper=12 * 20)
    s = Model (rate=0.065,
               balance=7000,
               taxrate=0.20)
    m.add_model (s)
    m.run ()
    #pprint (s.get_model_data ())
